// WavePlayer.cpp

#include "stdafx.h"
#include <stdexcept>
#include "WavePlayer.h"

namespace DDJCColumnWave	{

WavePlayer::WavePlayer(int b, int rate, int nc) : Wave(b, rate, nc, 4096)
{
	stopping = false;
	playing = false;
	paused = false;
	wavedevice = 0;
	if (waveOutOpen(&wavedevice, WAVE_MAPPER, &wfmt, 
			(DWORD)waveOutProc, (DWORD)this, CALLBACK_FUNCTION) == 0)
		if (waveOutPrepareHeader(wavedevice, &hdr1, sizeof(WAVEHDR)) == 0)
			if (waveOutPrepareHeader(wavedevice, &hdr2, sizeof(WAVEHDR)) == 0)
				return;
	Shutdown();
	throw runtime_error("Error initializing playback device");
}

WavePlayer::~WavePlayer()
{
	Shutdown();
}
void WavePlayer::Shutdown()
{
	waveOutUnprepareHeader(wavedevice, &hdr1, sizeof(WAVEHDR));
	waveOutUnprepareHeader(wavedevice, &hdr2, sizeof(WAVEHDR));
	do
		Sleep(50);
	while (waveOutClose(wavedevice) == WAVERR_STILLPLAYING);
}
void WavePlayer::Play()
{
	playing = true;
	FillBufferAndPlay(&hdr1);
	FillBufferAndPlay(&hdr2);
}

void WavePlayer::Stop()
{
	stopping = playing;
	if (paused)
		Resume();
}

void WavePlayer::Resume()
{
	if (paused)	{
		paused = false;
		Play();
	}
}

void WavePlayer::FillBufferAndPlay(PWAVEHDR hdr)
{
	if (playing)	{
		if (!stopping && !paused)	{
			if (hdr->dwFlags & WHDR_DONE)	{
				int len;
				len = FillBuffer(hdr->lpData, buflen);
				if (len > 0)
					len = DSP(hdr->lpData, len);
				if (len > buflen * 2)
					throw logic_error("FillBuffer/DSP returns too many samples");
				if (len > 0)	{
					hdr->dwBufferLength = len;
					if (waveOutWrite(wavedevice, hdr, sizeof(WAVEHDR)) != 0)
						throw runtime_error("Audio playback error");
				}
			}
		}
		else	{
			while((hdr1.dwFlags & WHDR_DONE) == 0 && 
					(hdr2.dwFlags & WHDR_DONE) == 0)
				Sleep(50);
			if (stopping)	{
				playing = false;
				paused = false;
			}
		}
	}
}

void CALLBACK WavePlayer::waveOutProc(HWAVEOUT hwo,UINT uMsg,DWORD 
									  dwInstance,DWORD dwParam1,DWORD dwParam2)
{
	WavePlayer* wp = reinterpret_cast<WavePlayer*>(dwInstance);
	PWAVEHDR hdr = reinterpret_cast<PWAVEHDR>(dwParam1);
	switch (uMsg)	{
		case WOM_DONE:
			wp->FillBufferAndPlay(hdr);
			break;
		default:
			break;
	}
}

} // namespace DDJCColumnWave
