/* bobcons.c - 'Cons' handler */
/*
        Copyright (c) 2001, by David Michael Betz
        All rights reserved
*/

#include "bob.h"

/* CONS */

#define BobConsP(o)                     BobIsType(o,BobConsDispatch)
#define BobCar(o)                       BobFixedVectorElement(o,0) 
#define BobSetCar(o,v)                  BobSetFixedVectorElement(o,0,v)
#define BobCdr(o)                       BobFixedVectorElement(o,1)
#define BobSetCdr(o,v)                  BobSetFixedVectorElement(o,1,v)
#define BobConsSize                     2

BobValue BobCons(BobInterpreter *c,BobValue car,BobValue cdr);
extern BobDispatch *BobConsDispatch;

BobDispatch *BobConsDispatch = NULL;

static BobValue BIF_List(BobInterpreter *c);
static BobValue BIF_Cons(BobInterpreter *c);
static BobValue BIF_Car(BobInterpreter *c);
static BobValue BIF_SetCar(BobInterpreter *c);
static BobValue BIF_Cdr(BobInterpreter *c);
static BobValue BIF_SetCdr(BobInterpreter *c);

/* function table */
static BobCMethod functionTable[] = {
BobMethodEntry( "List",             BIF_List            ),
BobMethodEntry( "Cons",             BIF_Cons            ),
BobMethodEntry( "Car",              BIF_Car             ),
BobMethodEntry( "SetCar",           BIF_SetCar          ),
BobMethodEntry( "Cdr",              BIF_Cdr             ),
BobMethodEntry( "SetCdr",           BIF_SetCdr          ),
BobMethodEntry( 0,					0					)
};

/* BobInitCons - initialize the 'Cons' type */
int BobInitCons(void)
{
    if (!(BobConsDispatch = BobMakeFixedVectorType(NULL,"Cons",BobConsSize)))
        return FALSE;
    return TRUE;
}

/* BobUseCons - initialize the 'Cons' object */
void BobUseCons(BobInterpreter *c)
{
    BobEnterFunctions(BobGlobalScope(c),functionTable);
}

/* BobCons - construct a pair */
BobValue BobCons(BobInterpreter *c,BobValue car,BobValue cdr)
{
    BobValue new;
    BobCheck(c,2);
    BobPush(c,cdr);
    BobPush(c,car);
    new = BobMakeFixedVectorValue(c,BobConsDispatch);
    BobSetCar(new,BobPop(c));
    BobSetCdr(new,BobPop(c));
    return new;
}

/* BIF_List - built-in function 'List' */
static BobValue BIF_List(BobInterpreter *c)
{
    int n = BobArgCnt(c);
    BobValue *p = BobArgPtr(c) - n;
    BobValue val = c->nilValue;
    while (--n >= 0)
        val = BobCons(c,*p++,val);
    return val;
}

/* BIF_Cons - built-in function 'Cons' */
static BobValue BIF_Cons(BobInterpreter *c)
{
    BobCheckArgCnt(c,2);
    return BobCons(c,BobGetArg(c,1),BobGetArg(c,2));
}

/* BIF_Car - built-in function 'Car' */
static BobValue BIF_Car(BobInterpreter *c)
{
    BobCheckArgCnt(c,1);
    BobCheckType(c,1,BobConsP);
    return BobCar(BobGetArg(c,1));
}

/* BIF_SetCar - built-in function 'SetCar' */
static BobValue BIF_SetCar(BobInterpreter *c)
{
    BobCheckArgCnt(c,2);
    BobCheckType(c,1,BobConsP);
    return BobSetCar(BobGetArg(c,1),BobGetArg(c,2));
}

/* BIF_Cdr - built-in function 'Cdr' */
static BobValue BIF_Cdr(BobInterpreter *c)
{
    BobCheckArgCnt(c,1);
    BobCheckType(c,1,BobConsP);
    return BobCdr(BobGetArg(c,1));
}

/* BIF_SetCdr - built-in function 'SetCdr' */
static BobValue BIF_SetCdr(BobInterpreter *c)
{
    BobCheckArgCnt(c,2);
    BobCheckType(c,1,BobConsP);
    return BobSetCdr(BobGetArg(c,1),BobGetArg(c,2));
}

