/* bobforeign.c - 'ForeignPointer' handler */
/*
        Copyright (c) 2001, by David Michael Betz
        All rights reserved
*/

#include <string.h>
#include "bob.h"

#define ForeignPtrNext(o)               (((BobForeignPtr *)o)->next)  
#define SetForeignPtrNext(o,v)          (((BobForeignPtr *)o)->next = (v))  
#define SetForeignPtr(o,v)              (((BobForeignPtr *)o)->value = (v))  

static int ForeignPtrPrint(BobInterpreter *c,BobValue obj,BobStream *s);
static long ForeignPtrSize(BobValue obj);
static void ForeignPtrScan(BobInterpreter *c,BobValue obj);

BobDispatch BobForeignPtrDispatch = {
    "ForeignPointer",
    BobDefaultGetProperty,
    BobDefaultSetProperty,
    BobDefaultNewInstance,
    ForeignPtrPrint,
    ForeignPtrSize,
    BobDefaultCopy,
    ForeignPtrScan,
    BobDefaultHash,
    FALSE
};

static int ForeignPtrPrint(BobInterpreter *c,BobValue obj,BobStream *s)
{
    char buf[64];
    sprintf(buf,"%08lx:%08lx",(long)obj,(long)BobForeignPtrValue(obj));
    return BobStreamPutC('<',s) == '<'
        && BobStreamPutS(BobTypeName(obj),s) == 0
        && BobStreamPutC('-',s) == '-'
        && BobStreamPutS(buf,s) == 0
        && BobStreamPutC('>',s) == '>';
}

static long ForeignPtrSize(BobValue obj)
{
    return sizeof(BobForeignPtr);
}

static void ForeignPtrScan(BobInterpreter *c,BobValue obj)
{
    SetForeignPtrNext(obj,c->newSpace->cObjects);
    c->newSpace->cObjects = obj;
}

/* BobMakeForeignPtrType - make a new foreign pointer type */
BobDispatch *BobMakeForeignPtrType(BobInterpreter *c,char *typeName)
{
    return BobMakeType(c,typeName,&BobForeignPtrDispatch);
}

/* BobMakeForeignPtr - make a new foreign pointer value */
BobValue BobMakeForeignPtr(BobInterpreter *c,BobDispatch *d,void *ptr)
{
    BobValue new;
    new = BobAllocate(c,sizeof(BobForeignPtr));
    BobSetDispatch(new,d);
    SetForeignPtrNext(new,c->newSpace->cObjects);
    c->newSpace->cObjects = new;
    SetForeignPtr(new,ptr);
    return new;
}

