// listener.cpp : Defines the class behaviors for the application.
//

#include "stdafx.h"
#include "listener.h"

#include "MainFrm.h"
#include "listenerDoc.h"
#include "listenerView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define HEAP_SIZE   (1024 * 1024)
#define STACK_SIZE  (64 * 1024)

/* prototypes */
static int CloseListenerStream(BobStream *s);
static int ListenerStreamGetC(BobStream *s);
static int ListenerStreamPutC(int ch,BobStream *s);
static void ErrorHandler(BobInterpreter *c,int code,va_list ap);

/* dispatch structure for null streams */
static BobStreamDispatch listenerDispatch = {
    CloseListenerStream,
    ListenerStreamGetC,
    ListenerStreamPutC
};

/* stream for listener */
static BobStream listenerStream = { &listenerDispatch };

/////////////////////////////////////////////////////////////////////////////
// CListenerApp

BEGIN_MESSAGE_MAP(CListenerApp, CWinApp)
	//{{AFX_MSG_MAP(CListenerApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_FILE_LOAD_OBJECT, OnFileLoadObject)
	ON_COMMAND(ID_FILE_LOAD_SOURCE, OnFileLoadSource)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	//ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	//ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
	// Standard print setup command
	ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CListenerApp construction

CListenerApp::CListenerApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CListenerApp object

CListenerApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CListenerApp initialization

BOOL CListenerApp::InitInstance()
{
	// Standard initialization
	// If you are not using these features and wish to reduce the size
	//  of your final executable, you should remove from the following
	//  the specific initialization routines you do not need.

#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	// Change the registry key under which our settings are stored.
	// TODO: You should modify this string to be something appropriate
	// such as the name of your company or organization.
	SetRegistryKey(_T("Local AppWizard-Generated Applications"));

	LoadStdProfileSettings();  // Load standard INI file options (including MRU)

	// Register the application's document templates.  Document templates
	//  serve as the connection between documents, frame windows and views.

	CSingleDocTemplate* pDocTemplate;
	pDocTemplate = new CSingleDocTemplate(
		IDR_MAINFRAME,
		RUNTIME_CLASS(CListenerDoc),
		RUNTIME_CLASS(CMainFrame),       // main SDI frame window
		RUNTIME_CLASS(CListenerView));
	AddDocTemplate(pDocTemplate);

	// Parse command line for standard shell commands, DDE, file open
	CCommandLineInfo cmdInfo;
	ParseCommandLine(cmdInfo);

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo))
		return FALSE;

	// The one and only window has been initialized, so show and update it.
	m_pMainWnd->ShowWindow(SW_SHOW);
	m_pMainWnd->UpdateWindow();

    /* make the interpreter */
    if (!(m_interpreter = BobMakeInterpreter(HEAP_SIZE,STACK_SIZE)))
        return 0;

    /* setup the error handler */
    m_interpreter->errorHandler = ErrorHandler;
    m_interpreter->standardInput = &listenerStream;
    m_interpreter->standardOutput = &listenerStream;
    m_interpreter->standardError = &listenerStream;

    /* add the library functions to the symbol table */
    BobEnterLibrarySymbols(m_interpreter);

    /* use the math routines */
    BobUseMath(m_interpreter);

    /* enable the use of eval */
    BobUseEval(m_interpreter);

	return TRUE;
}

int EvalString(BobInterpreter *i,char *str,int len)
{
	BobUnwindTarget target;
    int sts;

	/* setup the error handler target */
	BobPushUnwindTarget(i,&target);

	/* catch errors */
    if ((sts = BobUnwindCatch(i)) == 0) {
        BobStream *s = BobMakeStringStream((unsigned char *)str,len);
	    if (s) {
		    BobValue val;

		    /* evaluate the stream */
		    val = BobEvalStream(BobGlobalScope(i),s);

		    /* print the value of the expression */
		    BobStreamPutS("--> ",i->standardOutput);
		    BobPrint(i,val,i->standardOutput);
		    BobStreamPutC('\n',i->standardOutput);

		    /* close the stream */
		    BobCloseStream(s);
	    }
    }

	/* pop the unwind target */
	BobPopUnwindTarget(i);

    /* return status */
    return sts == 0;
}

/* CloseListenerStream - listener stream close handler */
static int CloseListenerStream(BobStream *s)
{
    return 0;
}

/* ListenerStreamGetC - listener stream getc handler */
static int ListenerStreamGetC(BobStream *s)
{
	return EOF;
}

/* ListenerStreamPutC - listener stream putc handler */
static int ListenerStreamPutC(int ch,BobStream *s)
{
    return theApp.m_transcriptView->PutC(ch);
}

/* ErrorHandler - error handler callback */
static void ErrorHandler(BobInterpreter *c,int code,va_list ap)
{
    switch (code) {
    case BobErrExit:
        PostQuitMessage(0);
        break;
    default:
        BobShowError(c,code,ap);
        BobStackTrace(c);
        break;
    }
    BobAbort(c);
}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CListenerApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
// CListenerApp message handlers


void CListenerApp::OnFileLoadSource() 
{
    CFileDialog dlg(TRUE,"bob",NULL,0,"Bob Source Files|*.bob|All Files|*.*||");
    BobInterpreter *i = theApp.m_interpreter;
    if (dlg.DoModal() == IDOK) {
        CString name = dlg.GetPathName();
		BobLoadFile(BobGlobalScope(i),(char *)(LPCTSTR)name,i->standardOutput);
    }
}

void CListenerApp::OnFileLoadObject() 
{
    CFileDialog dlg(TRUE,"bob",NULL,0,"Bob Object Files|*.bbo|All Files|*.*||");
    BobInterpreter *i = theApp.m_interpreter;
    if (dlg.DoModal() == IDOK) {
        CString name = dlg.GetPathName();
		BobLoadObjectFile(BobGlobalScope(i),(char *)(LPCTSTR)name);
    }
}
